/*
 * This is necessary for table cell contents to wrap
 * when using the sphinx_rtd_theme
 *
 * see https://github.com/readthedocs/sphinx_rtd_theme/issues/1246
 */
.wy-table-responsive table td, .wy-table-responsive table th {
    white-space: inherit;
}
